#!/usr/local/bin/ruby
=begin

=Name
rif - a ruby version of unifdef(1).

=Synopsis
rif [-Dsym[=val]] [-Usym] [-Ipath] src ...

=Description
rif deletes conditionals of cpp(1) from a source file and leave
other parts of code unchanged. Processed files are output to the
standard output. Unlike original unifdef, rif reads all the
included files, includes nested ones, and utilize defines in them.

=Options
: -Dsym[=val]
  defines a symbol sym. If `=val' is supplied, sym's value is set
  to val.
: -Usym
  undefines a symbol sym.
: -Ipath
  adds a path to include paths. In default, rif searches included
  files from /usr/include, /usr/local/include and the current
  directory.

=Bugs
* Coexistence of numeric and boolean symbols in one expression may
  causes exception.
* Macros with arguments in conditionals do not work correctly.
* Macro definitions with comments are not properly treated.
* Error checking features are needed, such as detection of
  unbalanced #if, #ifdef, #ifndef or #endif.

=end


IncludePaths = ["/usr/include", "/usr/local/include", "."]
DefinedSymbol = {}
IncludedFiles = []
$stdout.sync = true


class FalseClass
  def <(o); false; end
  def >(o); false; end
  def ==(o); false; end
  def <=(o); false; end
  def >=(o); false; end
end

class TrueClass
  def <(o); false; end
  def >(o); false; end
  def ==(o); false; end
  def <=(o); false; end
  def >=(o); false; end
end

def proc_cpp(fn, out = nil)
  IncludePaths.each do
    |path|
    path = File.join(path, fn)
    if File.exists?(path)
      fn = path
    end
  end

  return if IncludedFiles.include?(fn)

  IncludedFiles << fn

  File.open(fn) do
    |inf|
    ln = ""
    while inf.gets
      ln << $_
      case ln
      when /\\\Z/
	# continuous line
	next
      when /^#\s*include\s+["<]([^">]+)[">]/
	proc_cpp($1)
      when /^#\s*define\s+(\S+\([^)]+\))\s+(.*)/
	DefinedSymbol[$1] = $2
      when /^#\s*define\s+(\S+)\s+(.*)/
	DefinedSymbol[$1] = $2
      when /^#\s*define\s+(\S+)/
	DefinedSymbol[$1] = true
      when /^#\s*undef\s+(\S+)/
	DefinedSymbol[$1] = nil
      when /^#\s*ifdef\s+(\S+)/
	if DefinedSymbol[$1]
	  # skip only this #ifdef
	else
	  skip_clause(inf)
	end
	ln.replace ""
	next
      when /^#\s*ifndef\s+(\S+)/
	if not DefinedSymbol[$1]
	  # skip only this #ifndef
	else
	  skip_clause(inf)
	end
	ln.replace ""
	next
      when /^#\s*if\s+/
	exp = $'
	exp.gsub!(/defined\s*\(?(\w+)\)?/) { DefinedSymbol[$1] ? "true" : "false" }
	exp.gsub!(/\b([_a-zA-Z]\w+)\b/) { DefinedSymbol[$1] || "false" }
	exp.gsub!(/\b((?:0x)?\d+)\w\b/) { $1 }

	if eval(exp)
	  # skip only this #if
	else
	  skip_clause(inf)
	end
	ln.replace ""
	next
      when /^#\s*else/
	skip_clause(inf)
	ln.replace ""
	next
      when /^#\s*endif/
	ln.replace ""
	next
      else
	# ....?
      end
      puts ln if out
      ln.replace ""
    end
  end
end

def skip_clause(inf, lev = 1)
  elv = lev
  while ln = inf.gets
    case ln
    when /^#\s*if/
      lev += 1
    when /^#\s*endif/
      lev -= 1
    when /^#\s*else/
      break if elv == lev
    end
    break if lev == 0
  end
end


Src = []

ARGV.each do
  |arg|
  case arg
  when /^-I(.+)/
    IncludePaths.unshift $1
  when /^-D(\w+)=(.+)/
    DefinedSymbol[$1] = $2
  when /^-D(\w+)/
    DefinedSymbol[$1] = true
  when /^-U(\w+)/
    DefinedSymbol[$1] = nil
  else
    Src << arg
  end
end

Src.each do
  |src|
  proc_cpp src, true
end
